/****************************************************************************** 
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008. All Rights Reserved.
 *  
 * U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 ******************************************************************************/
 // Summary:
 //    The TableHeaderRenderer is an internal class to the TableViewer. It's responsible
 //    for constructing the user controls for the table. This is an abstract base class for 
 //    two concrete header renderer classes: one for the GRID mode, which has a table heading
 //    and the other for the LIST mode. This abstract base classes provides methods 
 //    to create common controls for both table modes. Clients of the TableViewer don't have the need
 //    to customize the TableViewer's header and controls and thus, it doesn't have 
 //    APIs to set/get a TableHeaderRenderer at this point.
 // 
 //    This base class provides methods to construct the table's paging and filtering controls.
 //
dojo.provide("com.ibm.asq.common.web.ui.internal.widgets.tableViewer.TableHeaderRenderer");

dojo.require("dojox.data.dom");

dojo.require("com.ibm.asq.common.web.ui.internal.view.common.ImageCache");
dojo.require("com.ibm.asq.common.web.ui.internal.util.MessageUtil");
dojo.require("com.ibm.asq.common.web.ui.internal.widgets.tableViewer.TableConstants");

dojo.require("com.ibm.asq.common.web.client.internal.Constants");

( function() {

var Constants = com.ibm.asq.common.web.client.internal.Constants;
var TableConstants = com.ibm.asq.common.web.ui.internal.widgets.tableViewer.TableConstants;
var ImageCache     = com.ibm.asq.common.web.ui.internal.view.common.ImageCache;
var FilterInputDelay = 800;

dojo.declare("com.ibm.asq.common.web.ui.internal.widgets.tableViewer.TableHeaderRenderer", null,
{
    // the parent viewer that employs this header renderer
    viewer: null,
    sortByColumn: -1,
    sortDir: TableConstants.SORT_DIR_ASCEND,
    
    // group controls
    groupSelect: null,
    groupByColumn: -1,
    
    // page controls
    pageCtrlPrev: null,
    pageCtrlInfo: null,
    pageCtrlNext: null,
    
    // filter controls
    filterInput: null,
    resetFilter: null,
    
    constructor: function(viewer)
    {
        this.viewer = viewer;
        this.pageCtrlPrev = [];
        this.pageCtrlInfo = [];
        this.pageCtrlNext = [];
        this.delayFilterTimer = null;
    },
    
    // Summary:
    //    Reset any states that the header may have cached. This is usually called when
    //    the table viewer's content has been changed and any old state may not be applicable.
    //
    resetStates: function()
    {
        // base class have no cached states, do nothing.
    },
    
    // Summary:
    //   Creates the TableViewer's controls. The TableViewer calls this renderer method
    //   to create the controls for its view.
    //
    createTableControls: function(pageRangeSelectNode, pageNode, footerPageNode, filterControlNode, groupControlNode)
    {
        
    },
    
    // Summary:
    //   Creates the TableViewer's header. The TableViewer calls this renderer method
    //   to create the header for the table.
    //
    createTableHeader: function(table)
    {
        
    },
    
    // Summary:
    //    Updates the paging control display. This is to show the current page information.
    //
    updatePageControlDisplay: function()
    {
        // don't need to update the paging display.
        if(!this.viewer.getIsShowingPagingControl())   
            return;
              
        var total = this.viewer.getContentItemCount();
        var rangeStart = this.viewer.getCurrentPageStartItem();
        var span = (rangeStart + this.viewer.getContentPageSize() + this.viewer.getHiddenRowsSize());
        var rangeEnd = (this.viewer.getContentPageSize() > -1) ? Math.min(total, span) : total;
        var msg;
        var offsetPos = (total > 0) ? 1 : 0;
        
        // if the viewer is doing filtering, we have to take into the account
        // of the items that were filtered out. The item shown may not be the total
        // page size, so we have to use the current counts for filtered and shown items.
        //
        if(this.viewer.getFilter() != null) {
            var displayTotal = this.viewer.getCurrentDisplayItemsCount();
            var start = (this.viewer.nItemsShown > 0) ? (rangeStart + offsetPos) : 0;
            var end = (start > 0) ? start + this.viewer.nItemsShown - 1 : 0;
            msg = MessageUtil.getMessage(MessageBundle.COMMON, "tableViewer_pageRangeFiltered", start, end, displayTotal, (total - displayTotal));
        }
        else {
            msg = MessageUtil.getMessage(MessageBundle.COMMON, "tableViewer_pageRange", (rangeStart + offsetPos), rangeEnd, total);
        }
        var hasPrevPage = this.viewer.getHasPreviousPage();
        var hasNextPage = this.viewer.getHasNextPage();
        
        for(var i = 0; i < this.pageCtrlPrev.length; i++) {
            this.pageCtrlPrev[i].disabled = !hasPrevPage;
            this.pageCtrlNext[i].disabled = !hasNextPage;
            
            this.pageCtrlPrev[i].className = (this.pageCtrlPrev[i].disabled) ? "action-link-disabled" : "action-link";
            this.pageCtrlNext[i].className = (this.pageCtrlNext[i].disabled) ? "action-link-disabled" : "action-link";
                   
            this.pageCtrlInfo[i].innerHTML = msg;
        }
    },
    
    _createPageRangeControl: function(node)
    {
        var provider = this.viewer.getContentProvider();
        var content = this.viewer.getContent();
        var rows = provider.getContentObjects(content);
        var count = rows.length;
        var itemsPerPage = [10, 25, 50, 75, 100];
        
		var pageLabel = document.createElement("label");		
		var pageLabelInfo = document.createElement("span");
		pageLabelInfo.style.display = "none";
		pageLabelInfo.appendChild(document.createTextNode(Constants.getMessage("labelPageRange")));
		pageLabel.appendChild(pageLabelInfo);
		pageLabel.setAttribute("for", "select-page-range");
		
		this.pageRangeSelect = document.createElement("select");
        this.pageRangeSelect.onchange = dojo.hitch(this, "_handlePageRangeChanged");
		this.pageRangeSelect.id="select-page-range";
        
        var pos = 0;
        var initSize = 0;
        
        for(var i = 0; i < itemsPerPage.length; i++) {
            if(count > itemsPerPage[i]) {
                var option = new Option(itemsPerPage[i] + "");
                option.value = itemsPerPage[i];
                this.pageRangeSelect.options[pos++] = option;
                
                // there's an user set size, retain it.
                if(this.selectedPageSize && this.selectedPageSize == itemsPerPage[i]) {
                    option.selected = true;
                    initSize = itemsPerPage[i];
                }      
            }
            else {
                break;
            }
        }
        
        // a large number of items can lead to performance problems in the web browser.
        // only provide the "Show All" option when the total number of rows is less
        // than the maximum number of items per page as specified in the itemsPerPage
        // array defined above.
        if (this.pageRangeSelect.options.length <  itemsPerPage.length) {
            var showAllOpt = new Option(MessageUtil.getMessage(MessageBundle.COMMON, "tableViewer_showAllItems"));
            showAllOpt.value = -1;
            this.pageRangeSelect.options[pos++] = showAllOpt;

            // there's an user set size, retain it.
            if(this.selectedPageSize && this.selectedPageSize == -1) {
                showAllOpt.selected = true;
                initSize = -1;
            }
        }
        
        // there's no user selected page size, show the first
        // page size.
        if (initSize == 0) {
            this.pageRangeSelect.options[0].selected = true;
            initSize = parseInt(this.pageRangeSelect.options[0].value);
        }
        if(this.pageRangeSelect.options.length == 1) {
            this.pageRangeSelect.disabled = true;
        }  		
		
		var lang = dojo.locale.split('-')[0];
		var RTL = {ar:true,fa:true,he:true,ur:true,yi:true,iw:true};
		if (RTL[lang]) {
			node.appendChild(document.createTextNode(MessageUtil.getMessage(MessageBundle.COMMON, "tableViewer_itemsPerPage")));
			node.appendChild(pageLabel);
			node.appendChild(this.pageRangeSelect);
		}
		else {
			node.appendChild(pageLabel);
			node.appendChild(this.pageRangeSelect);
			node.appendChild(document.createTextNode(MessageUtil.getMessage(MessageBundle.COMMON, "tableViewer_itemsPerPage")));
		}
        // initialize the viewer's page size to the first item.
        this.viewer.setContentPageSize(initSize);    
    },
    
    _handlePageRangeChanged: function()
    {
        var howMany = -1;
        
        for(var i = 0; i < this.pageRangeSelect.options.length; i++) {
            if(this.pageRangeSelect.options[i].selected) {
                howMany = parseInt(this.pageRangeSelect.options[i].value);
                break;
            }
        }
        this.viewer.setContentPageSize(howMany);   
        this.viewer.curPageItemStart = 0;

        // record the user set page size. Retain this setting for the 
        // table's content changed and the controls are being recreated.
        this.selectedPageSize = howMany;

        this.viewer._loadTableContentsWithProgress();
        this.updatePageControlDisplay();                   
    },
    
    _createPageControls: function(pageCtrl)
    {
        var pageCtrlPrev = document.createElement("button");
        var pageCtrlInfo = document.createElement("span");
        var pageCtrlNext = document.createElement("button");
        
        pageCtrlPrev.innerHTML = Constants.getMessage("Previous");
        pageCtrlNext.innerHTML = Constants.getMessage("Next");

        pageCtrlPrev.className = "action-link";
        pageCtrlNext.className = "action-link";
        
        dojo.connect(pageCtrlPrev, "onclick", this, "_handlePagePrev");
        dojo.connect(pageCtrlNext, "onclick", this, "_handlePageNext");
        
        pageCtrl.appendChild(pageCtrlPrev);
        pageCtrl.appendChild(pageCtrlInfo);
        pageCtrl.appendChild(pageCtrlNext);
        
        this.pageCtrlPrev.push(pageCtrlPrev);
        this.pageCtrlInfo.push(pageCtrlInfo);
        this.pageCtrlNext.push(pageCtrlNext);
    },
    
    _handlePagePrev: function()
    {
        if(this.viewer.getHasPreviousPage()) {
            this.viewer.performPreviousPage();
            this.updatePageControlDisplay();
        }
    },
    
    _handlePageNext: function()
    {
        if(this.viewer.getHasNextPage()) {
            this.viewer.performNextPage();
            this.updatePageControlDisplay();            
        }
    },

    _constructGroupControl: function(controlNode)
    {
        //console.debug("entering: _constructGroupControl");

        var provider = this.viewer.getContentProvider();
        var content = this.viewer.getContent();
        var nColumns = provider.getNumColumns(content);
        var hasGroupables = false;
        
        for (var i = 0; i < nColumns; i++) {
            if (provider.getIsColumnGroupable(content, i)) {
                hasGroupables = true;
                break;
            }    
        }
        if(!hasGroupables) {
            return;
        }
        
		var selectLabel = document.createElement("label");
		
        this.groupSelect = document.createElement("select");
        this.groupSelect.onchange = dojo.hitch(this, "_handleGroupBy");
		
        // create an empty selection
		var pos = 1;
		if(!this.viewer.hideUngrouped)
	    {
			this.groupSelect.options[0] = new Option(MessageUtil.getMessage(MessageBundle.COMMON, "tableViewer_ungrouped"));
			this.groupSelect.options[0].selected = true;
			this.groupSelect.options[0].value = -1;
		}else
			pos = 0;        
		
        for (var i = 0; i < nColumns; i++) {
            if (provider.getIsColumnGroupable(this.content, i)) {
                var heading = provider.getColumnHeading(content, i);
                var option = new Option(heading);
                option.value = i;
				if(this.viewer.groupByColumn && this.viewer.groupByColumn == i)
					option.selected = true
				
                this.groupSelect.options[pos++] = option;
            }    
        }
		selectLabel.appendChild(this.groupSelect)
		controlNode.appendChild(document.createTextNode(MessageUtil.getMessage(MessageBundle.COMMON, "tableViewer_groupBy") + " "));
        controlNode.appendChild(selectLabel);
    },

    _handleGroupBy: function()
    {
        var groupByCol = -1;
        
        for(var i = 0; i < this.groupSelect.options.length; i++) {
            if(this.groupSelect.options[i].selected) {
                groupByCol = parseInt(this.groupSelect.options[i].value);
                break;
            }
        }
        this.viewer.performGroupBy(groupByCol);
    },
    
    _constructFilterControls: function(node)
    {
        var container = node;
        
		var inputLabel = document.createElement("label");
		var inputLabelInfo = document.createElement("span");
		inputLabelInfo.style.display = "none";
		if (this.viewer.tableSummary) {
			inputLabelInfo.appendChild(document.createTextNode(this.viewer.tableSummary + " " + Constants.getMessage("tableViewer_filterInputLabel")));
		}
		inputLabel.appendChild(inputLabelInfo);
		inputLabel.setAttribute("for", "filter-input");
		
		this.filterInput = document.createElement("input");
        this.filterInput.type = "text";
        this.filterInput.size = 23;
        this.filterInput.id = "filter-input";
		
        this.filterInput.onkeyup = dojo.hitch(this, "_filterInputSet");
        this.filterInput.onfocus = dojo.hitch(this, "_filterInputFocus");
        this.filterInput.onblur = dojo.hitch(this, "_filterInputBlur");
		        
        this.resetFilter = document.createElement("button");
        var img = document.createElement("img");
        
        img.src = ImageCache.CLEAR_FILTER_ENABLED;
        this.resetFilter.appendChild(img);
        
        dojo.addClass(this.resetFilter, "action-link");
        dojo.addClass(this.resetFilter, "clear-filter");
		if (this.viewer.tableSummary) {
			this.resetFilter.name = this.viewer.tableSummary + " " + Constants.getMessage("tableViewer_filterClean");
		}
        this.resetFilter.onclick = dojo.hitch(this, "_resetFilter");
        container.appendChild(inputLabel);
		container.appendChild(this.filterInput);
        container.appendChild(this.resetFilter);

        this._setFilterInputInactive();
    },

    _resetFilter: function()
    {
        this.viewer.performFilterBy(null);
        this._setFilterInputInactive();
    },

    _filterInputSet: function()
    {
        // The user has typed more chars, cancel the current
        // start of filtering and wait for more chars.
        //
        if(this.delayFilterTimer) {
            clearTimeout(this.delayFilterTimer);
            this.delayFilterTimer = null;
        }
        
        // set a small delay before doing the filtering.
        // If the user typed more chars, we want to queue up
        // more filter chars. Otherwise, we would be performing 
        // filtering on every char.
        this.delayFilterTimer = setTimeout(dojo.hitch(this, "_doFilter"), FilterInputDelay);        
    },
   
    _doFilter: function()
    {   
        this.delayFilterTimer = null;
        this.filterText = this.filterInput.value;
            
        // make it a null to indicate clearing of the filter.
        if(this.filterText && this.filterText.length == 0)
            this.filterText = null;
        
        this.viewer.performFilterBy(this.filterText);
    },
    
    _filterInputFocus: function()
    {
        this.filterInput.className = "";
        var hasFilter = this.viewer.getFilter() != null;
        this.filterInput.className = "filter-text-input";
        
        // we don't have a filter in the viewer. Clear the prompt
        // we have displayed in the input field.
        if(!hasFilter)
            this.filterInput.value = "";        
        
    },
    
    _filterInputBlur: function()
    {
        this._setFilterInputInactive();
    },
    
    _setFilterInputInactive: function()
    {
        this.filterInput.className = "filter-text-input-disabled";
        var noFilter = this.viewer.getFilter() == null;
        
        // we don't have a filter in the viewer, show the
        // default prompt in the filter field.
        if(noFilter)
            this.filterInput.value = MessageUtil.getMessage(MessageBundle.COMMON, "tableViewer_filterPrompt");   
    }
    
});

})();
