/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008. All Rights Reserved.
 *  
 * U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 ******************************************************************************/
package com.ibm.rational.test.lm.spi.bf.internal;

import java.io.IOException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.NoSuchAlgorithmException;
import java.util.HashMap;
import java.util.Stack;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;

import com.buildforge.services.client.api.APIClientConnection;
import com.buildforge.services.client.dbo.User;
import com.buildforge.services.common.ServiceException;
import com.buildforge.services.common.api.APIConstants;
import com.buildforge.services.common.api.APIException;
import com.buildforge.services.common.api.APIRequest;
import com.buildforge.services.common.text.Hex;
import com.buildforge.services.common.text.StringConverter;
import com.buildforge.services.common.text.TextUtils;
import com.ibm.rational.test.lm.spi.bf.common.BFAPIAuthenticationException;
import com.ibm.rational.test.lm.spi.bf.common.BFAPIClientConnectionException;
import com.ibm.rational.test.lm.spi.bf.common.BFAPIException;

/**
 * This class wraps the BuildForge API Connection. Multiple BF connections can
 * be constructed in parallel threads, as long as they share the same
 * authentication. If the same user is authenticated for a second time using
 * authUser(), the original connection is invalidated!
 * 
 * So we reference count the authentication (to be used across multiple threads)
 * and also reference count each connection. Using one BFAPIConnection, we allow
 * multiple pairs of connect/close calls. Each call increments/decrements a
 * counter, and the underlying connection is constructed once and then closed
 * when the last close() is called (when the counter goes to 0).
 * 
 * In addition, the authentication is cached statically per server/user, as long
 * as any instance of the BFAPIClientConnection is connected for that
 * server/user combination.
 * 
 * I believe that the expected pattern for connect/getConnection/close calls is
 * that they should only be called within a single thread for a single
 * JobController. If this is true, it is probably sufficient to synchronize
 * access to the token object itself, not to the entire connect/close method,
 * and the getConnection call probably need not be synchronized at all.
 * 
 * Note also that each process must have its own BF user; when we scale up to
 * multiple equinox processes, we will need to either configure multiple BF
 * users (one per process) or alternatively (suggested by Matt) we could put a
 * single equinox process in front of the single BF process that we are using.
 * 
 * 
 * @author FMcGrath
 * 
 */
public class BFAPIClientConnection {
	private static final Class<BFAPIClientConnection> CLASS = BFAPIClientConnection.class;
	// static final ITracer tracer =
	// RlmTracer.getTracer(SPIBFActivator.PLUGIN_ID);

	private APIClientConnection conn = null;

	private String hostname = null;
	private int port = 0;
	private String server = null;
	private String hostSystemPassword = null;

	private String domain = null; // if using LDAP user auth.
	private String user = null;
	private int userNumber = -1;
	private String password = null;
	private int connectionCount = 0;
	String authKey = null;

	/**
	 * This manages the caching of authentication for multiple users.
	 * 
	 * @author FMcGrath
	 * 
	 */
	private static class AuthInfo {
		public String token;
		public int authCount;

		public AuthInfo(String token) {
			this.token = token;
			this.authCount = 1;
		}
	}

	/**
	 * For instances of buildforge where we do not authenticate with a system
	 * account we keep track of the open connections for the specified user, and
	 * re-use the authenication if another connection is made with the same
	 * server/user.
	 */
	private static HashMap<String, AuthInfo> authorizations = new HashMap<String, AuthInfo>();

	/**
	 * This maps Strings designating servers by hostname:port to a buildforge
	 * API connection.
	 */
	private static HashMap<String, Stack<APIClientConnection>> serversToApiConnections = new HashMap<String, Stack<APIClientConnection>>();

	private static APIClientConnection getCachedSystemConnection(String server) {
		Stack<APIClientConnection> v = serversToApiConnections.get(server);
		if (v == null)
			return null;
		if (v.empty())
			return null;
		String test = "test";
		APIClientConnection conn = v.pop();
		try {
			// Test if the connection still alive
			// If the connection are not alive, we need to reconnect one
			if (conn.ping(test) == null) {
				conn = null;
			}
		} catch (IOException e) {
			return null;
		} catch (ServiceException e) {
			// TODO Auto-generated catch block
			return null;
		}
		return conn;

	}

	private static void releaseSystemConnection(String server,
			APIClientConnection connection) {
		final String methName = "releaseSystemConnection"; //$NON-NLS-1$
		try {
			connection.check();
		} catch (Exception e) {
			// tracer.trace(TraceType.NOTE, Level.FINE, CLASS, methName,
			// "BuildForge APIClientConnection was dropped", e); //$NON-NLS-1$
			// tracer.error(CLASS, methName,
			// Messages.BUILDFORGE_CONNECTION_DROP_ERROR, e);
			return;
		}
		Stack<APIClientConnection> v = serversToApiConnections.get(server);
		if (v == null) {
			v = new Stack<APIClientConnection>();
			serversToApiConnections.put(server, v);
		}
		connection.as(APIConstants.UID_SYSTEM);
		v.push(connection);
	}

	/**
	 * Access the authentication token for this key. If it is found, increment
	 * its reference count.
	 * 
	 * @param authKey
	 * @return
	 */
	private static String getAuthToken(String authKey) {
		AuthInfo ai = authorizations.get(authKey);
		if (ai == null)
			return null;

		++ai.authCount;
		return ai.token;
	}

	/**
	 * register a token for the authentication key.
	 * 
	 * @param authKey
	 * @param token
	 */
	private static void registerAuth(String authKey, String token) {
		AuthInfo ai = new AuthInfo(token);
		authorizations.put(authKey, ai);
	}

	/**
	 * release an authentication key. If the reference goes to 0, delete the
	 * token from the cache.
	 * 
	 * @param authKey
	 */
	private static void ungetAuth(String authKey) {
		AuthInfo ai = authorizations.get(authKey);
		--ai.authCount;
		if (ai.authCount == 0) {
			authorizations.remove(authKey);
		}
	}

	/**
	 * Construct the authentication key from the relevant information.
	 * 
	 * @param hostname
	 * @param port
	 * @param user
	 * @return
	 */
	private static String getAuthKey(String hostname, int port, String user) {
		StringBuilder sb = new StringBuilder(hostname);
		sb.append(':');
		sb.append(Integer.toString(port));
		sb.append(':');
		sb.append(user);
		return sb.toString();
	}

	/**
	 * Constructor - does not do anything active. By default, BF uses
	 * localhost:3966
	 * 
	 * @param user -
	 *            the user to use for authentication.
	 * @param password -
	 *            the passeword
	 */
	public BFAPIClientConnection(String hostname, int port, String domain,
			String user, String password) {
		this.hostname = hostname;
		this.hostSystemPassword = null;
		this.port = port;
		this.server = new StringBuilder(hostname).append(':').append(port)
				.toString();
		this.domain = domain;
		this.user = user;
		this.userNumber = -1;
		this.password = password;
		this.authKey = getAuthKey(hostname, port, user);
	}

	/**
	 * Constructor - does not do anything active. By default, BF uses
	 * localhost:3966
	 * 
	 * @param user -
	 *            the user to use for authentication.
	 * @param password -
	 *            the passeword
	 */
	public BFAPIClientConnection(String hostname, int port, String domain,
			String hostSystemPassword) {
		this.hostname = hostname;
		this.hostSystemPassword = hostSystemPassword;
		this.port = port;
		this.server = new StringBuilder(hostname).append(':').append(port)
				.toString();
		this.domain = domain;
		this.user = null;
		this.userNumber = -1;
		this.password = null;
		this.authKey = getAuthKey(hostname, port, user);
	}

	public synchronized APIClientConnection connect()
			throws BFAPIException {
		return (this.hostSystemPassword != null) ? connectAsSystem()
				: connectAsUser();
	}

	/**
	 * Constructs and returns a BuildForge APIClientConnection. Caches it so
	 * that subsequent calls to connect() use the same APIClientConnection. In
	 * addition, caches the authentication token for this user so that if any
	 * other APIClientConnection objects are already connected, this uses the
	 * shared, cached authentication token.
	 * 
	 * @return The buildForge APIClientConnection
	 */
	public APIClientConnection connectAsUser()
			throws BFAPIClientConnectionException, BFAPIAuthenticationException {
		final String methName = "connectAsUser"; //$NON-NLS-1$
		if (conn != null) {
			// we already have an API Connection, just use it!
			assert (connectionCount > 0);
			++connectionCount;
			return conn;
		}
		try {

			conn = new APIClientConnection(this.hostname, this.port);
		} catch (Exception e) {
			connectionCount = 0;
			// internal
			BFAPIClientConnectionException bfce = new BFAPIClientConnectionException(
					"cannot construct APIClientConnection"); //$NON-NLS-1$
			bfce.initCause(e);
			// tracer.trace(TraceType.NOTE, Level.FINER, CLASS, methName,
			// "cannot construct APIClientConnection", bfce); //$NON-NLS-1$
			throw bfce;
		}

		try {

			String token = getAuthToken(authKey);
			if (token == null) {
				// we do not have cached authentication for this bfserver/user.
				token = conn.authUser(user, password, domain);
				registerAuth(authKey, token);
			} else {
				conn.authToken(token);
			}
		} catch (IOException e) {
			try {
				conn.close();
			} catch (IOException e2) { /* TODO: handle this exception */
			}
			BFAPIClientConnectionException bfce = new BFAPIClientConnectionException(
					"cannot authenticate APIClientConnection"); //$NON-NLS-1$
			bfce.initCause(e);
			// tracer.trace(TraceType.NOTE, Level.FINER, CLASS, methName,
			// "cannot authenticate APIClientConnection", bfce); //$NON-NLS-1$
			throw bfce;
		} catch (ServiceException e) {
			try {
				conn.close();
			} catch (IOException e2) { /* TODO: handle this exception */
			}
			BFAPIAuthenticationException bfae = new BFAPIAuthenticationException(
					"cannot authenticate APIClientConnection"); //$NON-NLS-1$
			bfae.initCause(e);
			// tracer.trace(TraceType.NOTE, Level.FINEST, CLASS, methName,
			// "cannot authenticate APIClientConnection", bfae); //$NON-NLS-1$
			throw bfae;
		}

		++connectionCount;
		return conn;
	}

	public void setCurrentUser(String user) {
		if (user == null)
			return;
		assert (conn == null);
		this.user = user;
		this.userNumber = -1;
	}

	/**
	 * Returns a BuildForge APIClientConnection. Caches it so that subsequent
	 * calls to connect() use the same APIClientConnection.
	 * 
	 * @return The buildForge APIClientConnection
	 */
	public APIClientConnection connectAsSystem()
			throws BFAPIException{
		final String methName = "connectAsSystem"; //$NON-NLS-1$
		if (conn != null) {
			// we already have an API Connection, just use it!
			assert (connectionCount > 0);
			++connectionCount;
			return conn;
		}
		conn = getCachedSystemConnection(server);
		if (conn == null) {
			try {
				conn = new APIClientConnection(this.hostname, this.port);
				// We had problems using JSON (BF bug) and it did not
				// work with the SYSTEM auth when changing the user. Anyway
				// JSON is now default and should work.
				// conn.setProtocol(Protocol.JSON);
				// conn.setProtocol(Protocol.BINARY);
			} catch (Exception e) {
				connectionCount = 0;
				BFAPIClientConnectionException bfce = new BFAPIClientConnectionException(
						"cannot construct APIClientConnection"); //$NON-NLS-1$
				bfce.initCause(e);
				// tracer.trace(TraceType.NOTE, Level.FINER, CLASS, methName,
				// "cannot construct APIClientConnection", bfce); //$NON-NLS-1$
				throw bfce;
			}
			try {
				System.out.println(">>>Connecting as SYSTEM"); //$NON-NLS-1$
				doSystemLogin(conn, hostSystemPassword);
				// we do not have cached authentication for this bfserver/user.
				// token = conn.authUser(user, password);
				// registerAuth(authKey,token);

			} catch (APIException e){
				final String msg = "the build forge API version is not matched";//$NON-NLS-1$
				BFAPIException bfce = new BFAPIException(
						msg); //$NON-NLS-1$
				bfce.initCause(e);
				throw bfce;
			} catch (IOException e) {
				try {
					conn.close();
				} catch (IOException e2) { /* TODO: handle this exception */
				}
				final String msg = "cannot create System connection APIClientConnection";//$NON-NLS-1$
				BFAPIClientConnectionException bfce = new BFAPIClientConnectionException(
						msg); //$NON-NLS-1$
				bfce.initCause(e);
				// tracer.trace(TraceType.NOTE, Level.FINE, CLASS, methName,
				// msg,
				// bfce);
				// tracer.error(CLASS, methName,
				// Messages.BUILDFORGE_CONNECTION_INIT_ERROR, bfce);
				throw bfce;
			} catch (Exception e) {
				try {
					conn.close();
				} catch (IOException e2) { /* TODO: handle this exception */
				}				
				final String msg = "cannot authenticate APIClientConnection"; //$NON-NLS-1$
				BFAPIAuthenticationException bfae = new BFAPIAuthenticationException(
						msg);
				bfae.initCause(e);
				throw bfae;
			}
		}
		// if we were given a user, then we want to switch to it;
		// if not, we operate as the system user.
		if (user != null) {
			if (userNumber == -1) {
				User u = null;
				try {
					u = User.findByLogin(conn, user);
				} catch (Exception e) {
					releaseSystemConnection(server, conn);
					conn = null;
					BFAPIClientConnectionException bfce = new BFAPIClientConnectionException(
							"cannot find user for APIClientConnection"); //$NON-NLS-1$
					bfce.initCause(e);
					// tracer
					// .trace(
					// TraceType.NOTE,
					// Level.FINEST,
					// CLASS,
					// methName,
					// "cannot find user for APIClientConnection: " + user,
					// bfce); //$NON-NLS-1$
					throw bfce;
				}
				userNumber = u.getUserId();
			}
			conn.as(userNumber);
		}
		++connectionCount;
		return conn;
	}

	/**
	 * Returns the BuildForge APIClientConnection.
	 * 
	 * @return the BuildForge APIClientConnection.
	 */
	public APIClientConnection getConnection()
			throws BFAPIClientConnectionException {
		final String methName = "getConnection"; //$NON-NLS-1$
		if (conn == null) {
			BFAPIClientConnectionException bfce = new BFAPIClientConnectionException(
					"getConnection called without connect"); //$NON-NLS-1$
			// tracer.trace(TraceType.NOTE, Level.FINE, CLASS, methName,
			// "getConnection called without connect", bfce); //$NON-NLS-1$
			throw bfce;
		}
		return conn;
	}

	/**
	 * closes the BuildForge APIClientConnection (if the reference count goes to
	 * 0). Also deletes the authentication token if this is the last connection.
	 * 
	 */
	public synchronized void close() throws BFAPIClientConnectionException {
		final String methName = "close"; //$NON-NLS-1$
		if (connectionCount <= 0) {
			BFAPIClientConnectionException bfce = new BFAPIClientConnectionException(
					"APIClientConnection.close() called when not connected"); //$NON-NLS-1$
			// tracer
			// .trace(
			// TraceType.NOTE,
			// Level.FINER,
			// CLASS,
			// methName,
			// "APIClientConnection.close() called when not connected", bfce);
			// //$NON-NLS-1$
			throw bfce;
		}
		if (--connectionCount != 0)
			return;

		if (hostSystemPassword != null) {
			releaseSystemConnection(server, conn);
			conn = null;
			return;
		}

		BFAPIClientConnectionException jce = null;
		try {
			conn.logout();
		} catch (Exception e) {
			jce = new BFAPIClientConnectionException(
					"APIClientConnection.logout error"); //$NON-NLS-1$
			jce.initCause(e);
			// tracer.trace(TraceType.NOTE, Level.FINER, CLASS, methName,
			// "APIClientConnection.logout error", jce); //$NON-NLS-1$
		}
		// try to close even if the logout failed
		try {
			conn.close();
		} catch (Exception e) {
			// close error more important?
			jce = new BFAPIClientConnectionException(
					"APIClientConnection.close error"); //$NON-NLS-1$
			jce.initCause(e);
			// tracer.trace(TraceType.NOTE, Level.FINER, CLASS, methName,
			// "APIClientConnection.close error", jce); //$NON-NLS-1$
		}

		conn = null;
		ungetAuth(authKey);
		if (jce != null)
			throw jce;
	}

	// note - we are going to switch to authenticating as System. To
	// do this we will need a system password, which is created by
	// adding a db_system_password setting to buildforge.conf

	private static final String HASH_ALG = "HMacMD5"; //$NON-NLS-1$

	private void doSystemLogin(APIClientConnection conn, String password)
			throws IOException, ServiceException, NoSuchAlgorithmException,
			InvalidKeyException {
		StringConverter sc = StringConverter.get();

		// These are magic strings you have to know to auth as SYSTEM. This
		// is to prevent people from using SYSTEM for everyday stuff. It
		// should never be made public knowledge.
		String seed = "?KWBnGs]4TLMY6W<g0O+TT&e"; //$NON-NLS-1$
		String salt = "~Wi18=w!@!5aCd2,4R%.&E[F"; //$NON-NLS-1$

		conn.request(APIConstants.COMMAND_LOGIN);
		conn.writeEntry("seed", seed); //$NON-NLS-1$
		APIRequest resp = conn.call();

		String challenge = resp.getString("challenge"); //$NON-NLS-1$
		if (TextUtils.isEmpty(challenge)) {
			throw APIException.missing("challenge");} //$NON-NLS-1$
		byte[] msgBytes = Hex.hexToBytes(sc.toLatin1(challenge));

		String secret = salt + password;
		byte[] keyBytes = sc.toUtf8Bytes(secret);

		Key key = new SecretKeySpec(keyBytes, 0, keyBytes.length, HASH_ALG);
		Mac mac = Mac.getInstance(HASH_ALG);
		mac.init(key);
		byte[] hash = mac.doFinal(msgBytes);
		String response = sc.fromLatin1(Hex.bytesToHex(hash));

		conn.request(APIConstants.COMMAND_LOGIN);
		conn.writeEntry("response", response); //$NON-NLS-1$
		conn.call();
	}

	/**
	 * Get the Build Forge connection and try to connect.
	 * 
	 * @return Build Forge Client connection
	 * @throws ServiceException
	 * @throws IOException
	 * @throws BFAPIAuthenticationException
	 * @throws BFAPIClientConnectionException
	 * @throws BFAPIException
	 */
	public static BFAPIClientConnection getConnection(
			String provider_server_name, int provider_server_port)
			throws IOException, ServiceException,
			BFAPIException {
		BFAPIClientConnection connection = ConnectionController
				.getSystemConnection(provider_server_name, provider_server_port);
		connection.connect();
		return connection;
	}
}
